#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
from PIL import Image
import tempfile
import os
from tensorflow.keras.callbacks import CSVLogger
#Internal libraries
from aucmedi.utils.callbacks import *
from aucmedi.utils.visualizer import *
from aucmedi import *

#-----------------------------------------------------#
#                  Unittest: Utility                  #
#-----------------------------------------------------#
class UtilityTEST(unittest.TestCase):
    # Create random imaging data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Initialize temporary directory
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".data")
        # Create Grayscale data for 2D
        self.img_gray_2D = np.random.rand(16, 16, 1) * 255
        # Create RGB data for 2D
        self.sampleList_rgb_2D = []
        for i in range(0, 5):
            img_rgb = np.random.rand(16, 16, 3) * 255
            imgRGB_pillow = Image.fromarray(img_rgb.astype(np.uint8))
            index = "image.sample_" + str(i) + ".RGB.png"
            path_sampleRGB = os.path.join(self.tmp_data.name, index)
            imgRGB_pillow.save(path_sampleRGB)
            self.sampleList_rgb_2D.append(index)
        self.img_rgb_2D = np.random.rand(16, 16, 3) * 255
        # Create Grayscale data for 3D
        self.img_gray_3D = np.random.rand(16, 16, 16, 1) * 255
        # Create HU data for 3D
        self.img_hu_3D = (np.random.rand(16, 16, 16, 1) * 1000) - 500
        # Create RGB data for 3D
        self.img_rgb_3D = np.random.rand(16, 16, 16, 3) * 255
        # Create classification labels
        self.labels_ohe = np.zeros((5, 4), dtype=np.uint8)
        for i in range(0, 5):
            class_index = np.random.randint(0, 4)
            self.labels_ohe[i][class_index] = 1
        # Create xai heatmap
        self.heatmap = np.random.rand(16, 16)
        # Create RGB Data Generator
        self.datagen = DataGenerator(self.sampleList_rgb_2D,
                                     self.tmp_data.name,
                                     labels=self.labels_ohe,
                                     resize=(16, 16),
                                     grayscale=False, batch_size=1)

    #-------------------------------------------------#
    #             Callbacks: CSV2history              #
    #-------------------------------------------------#
    def test_Callbacks_csv2history(self):
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".model")
        path_csv = os.path.join(self.tmp_data.name, "testing.csv")
        csvlog = CSVLogger(path_csv)

        model = NeuralNetwork(n_labels=4, channels=3, input_shape=(16,16))
        hist_returned = model.train(training_generator=self.datagen,
                                    validation_generator=self.datagen,
                                    epochs=3, callbacks=[csvlog])

        hist_loaded = csv_to_history(path_csv)
        del hist_loaded["epoch"]

        for key in hist_returned:
            self.assertTrue(key in hist_returned and key in hist_loaded)
            self.assertTrue(len(hist_returned[key]) == len(hist_loaded[key]))

    #-------------------------------------------------#
    #                Visualizer: Image                #
    #-------------------------------------------------#
    def test_Visualizer_Image_grayscale(self):
        # PNG
        path_out = os.path.join(self.tmp_data.name, "viz.image.2D.gray.png")
        visualize_image(self.img_gray_2D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # JPG
        path_out = os.path.join(self.tmp_data.name, "viz.image.2D.gray.jpg")
        visualize_image(self.img_gray_2D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # TIF
        path_out = os.path.join(self.tmp_data.name, "viz.image.2D.gray.tif")
        visualize_image(self.img_gray_2D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))

    def test_Visualizer_Image_rgb(self):
        # PNG
        path_out = os.path.join(self.tmp_data.name, "viz.image.rgb.png")
        visualize_image(self.img_rgb_2D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # JPG
        path_out = os.path.join(self.tmp_data.name, "viz.image.rgb.jpg")
        visualize_image(self.img_rgb_2D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # TIF
        path_out = os.path.join(self.tmp_data.name, "viz.image.rgb.tif")
        visualize_image(self.img_rgb_2D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))

    #-------------------------------------------------#
    #                Visualizer: Volume               #
    #-------------------------------------------------#
    def test_Visualizer_Volume_grayscale(self):
        # NumPy
        path_out = os.path.join(self.tmp_data.name, "viz.volume.gray.npy")
        visualize_volume(self.img_gray_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # MHA
        path_out = os.path.join(self.tmp_data.name, "viz.volume.gray.mha")
        visualize_volume(self.img_gray_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # NIfTI
        path_out = os.path.join(self.tmp_data.name, "viz.image.gray.nii")
        visualize_volume(self.img_gray_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # GIF
        path_out = os.path.join(self.tmp_data.name, "viz.image.gray.gif")
        visualize_volume(self.img_gray_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))

    def test_Visualizer_Volume_HU(self):
        # NumPy
        path_out = os.path.join(self.tmp_data.name, "viz.volume.hu.npy")
        visualize_volume(self.img_hu_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # MHA
        path_out = os.path.join(self.tmp_data.name, "viz.volume.hu.mha")
        visualize_volume(self.img_hu_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # NIfTI
        path_out = os.path.join(self.tmp_data.name, "viz.image.hu.nii")
        visualize_volume(self.img_hu_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # GIF
        path_out = os.path.join(self.tmp_data.name, "viz.image.hu.gif")
        visualize_volume(self.img_hu_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))

    def test_Visualizer_Volume_rgb(self):
        # NumPy
        path_out = os.path.join(self.tmp_data.name, "viz.volume.rgb.npy")
        visualize_volume(self.img_rgb_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # MHA
        path_out = os.path.join(self.tmp_data.name, "viz.volume.rgb.mha")
        visualize_volume(self.img_rgb_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # NIfTI
        path_out = os.path.join(self.tmp_data.name, "viz.image.rgb.nii")
        visualize_volume(self.img_rgb_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))
        # GIF
        path_out = os.path.join(self.tmp_data.name, "viz.image.rgb.gif")
        visualize_volume(self.img_rgb_3D, out_path=path_out)
        self.assertTrue(os.path.exists(path_out))

    #-------------------------------------------------#
    #             Visualizer: XAI Heatmap             #
    #-------------------------------------------------#
    def test_Visualizer_Heatmap_2D_grayscale(self):
        # Default
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.2D.gray.png")
        visualize_heatmap(self.img_gray_2D, self.heatmap, 
                          overlay=True, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))
        os.remove(path_xai)
        self.assertTrue(not os.path.exists(path_xai))
        # With Overlay
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.2D.gray.png")
        visualize_heatmap(self.img_gray_2D, self.heatmap, 
                          overlay=False, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))

    def test_Visualizer_Heatmap_2D_rgb(self):
        # Default
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.2D.rgb.png")
        visualize_heatmap(self.img_rgb_2D, self.heatmap, 
                          overlay=True, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))
        os.remove(path_xai)
        self.assertTrue(not os.path.exists(path_xai))
        # With Overlay
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.2D.rgb.png")
        visualize_heatmap(self.img_rgb_2D, self.heatmap, 
                          overlay=False, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))

    def test_Visualizer_Heatmap_3D_gray(self):
        # Default
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.3D.gray.npy")
        visualize_heatmap(self.img_gray_3D, self.heatmap, 
                          overlay=True, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))
        os.remove(path_xai)
        self.assertTrue(not os.path.exists(path_xai))
        # With Overlay
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.3D.gray.npy")
        visualize_heatmap(self.img_gray_3D, self.heatmap, 
                          overlay=False, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))

    def test_Visualizer_Heatmap_3D_rgb(self):
        # Default
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.3D.rgb.npy")
        visualize_heatmap(self.img_rgb_3D, self.heatmap, 
                          overlay=True, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))
        os.remove(path_xai)
        self.assertTrue(not os.path.exists(path_xai))
        # With Overlay
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.3D.rgb.npy")
        visualize_heatmap(self.img_rgb_3D, self.heatmap, 
                          overlay=False, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))

    def test_Visualizer_Heatmap_3D_hu(self):
        # Default
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.3D.hu.npy")
        visualize_heatmap(self.img_hu_3D, self.heatmap, 
                          overlay=True, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))
        os.remove(path_xai)
        self.assertTrue(not os.path.exists(path_xai))
        # With Overlay
        path_xai = os.path.join(self.tmp_data.name, "viz.heatmap.3D.hu.npy")
        visualize_heatmap(self.img_hu_3D, self.heatmap, 
                          overlay=False, out_path=path_xai)
        self.assertTrue(os.path.exists(path_xai))





#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import numpy as np
from PIL import Image
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import SimpleITK as sitk
# Internal libraries
from aucmedi.data_processing.subfunctions import Standardize

#-----------------------------------------------------#
#                   Image Visualizer                  #
#-----------------------------------------------------#
def visualize_image(array, out_path=None):
    """ Simple wrapper function to visualize a NumPy matrix as image via PIL.

    ???+ info
        NumPy array shape has to be (x, y, channel) like this: (224, 224, 3)

    Args:
        array (numpy.ndarray):          NumPy matrix containing an image.
        out_path (str):                 Path in which image is stored (else live output).
    """
    # Ensure integer intensity values
    array = np.uint8(array)
    # Remove channel axis if grayscale
    if array.shape[-1] == 1 : array = np.reshape(array, array.shape[:-1])
    # Convert array to PIL image
    image = Image.fromarray(array)
    # Visualize or store image
    if out_path is None : image.show()
    else : image.save(out_path)

#-----------------------------------------------------#
#                  Volume Visualizer                  #
#-----------------------------------------------------#
def visualize_volume(array, out_path=None, iteration_axis=1):
    """ Simple wrapper function to visualize/store a NumPy matrix as volume.

    ???+ info
        NumPy array shape has to be (x, y, z, channel) like this: (128, 128, 128, 1)

    Args:
        array (numpy.ndarray):          NumPy matrix containing an image.
        out_path (str):                 Path in which volume is stored (else live output).
    """
    # Identify output format
    if out_path is None : format = "gif"
    elif out_path.split(".")[-1].lower() in ["gif", "mha", "nii", "gz", "npy"]:
        format = out_path.split(".")[-1]
    else : raise ValueError("Visualizer does not support image format!")

    # Encode as GIF
    if format == "gif":
        # Grayscale normalization if required
        if array.shape[-1] == 1 and (np.min(array) < 0 or np.max(array) > 255):
            array = Standardize(mode="grayscale").transform(array)
        elif array.shape[-1] != 1 and (np.min(array) < 0 or np.max(array) > 255):
            raise ValueError("Visualizer does not support multi-channel " + \
                            "non-RGB formats." + \
                            " Array min: " + str(np.min(array)) + \
                            " Array max: " + str(np.max(array))) 
        # Ensure integer intensity values
        array = np.uint8(array)

        # Create a figure and two axes objects from matplot
        fig = plt.figure()
        img = plt.imshow(np.take(array, 0, axis=iteration_axis),
                        cmap='gray', vmin=0, vmax=255, animated=True)
        # Update function to show the slice for the current frame
        def update(i):
            plt.suptitle("Slice: " + str(i))
            img.set_data(np.take(array, i, axis=iteration_axis))
            return img
        # Compute the animation (gif)
        ani = animation.FuncAnimation(fig, update, 
                                    frames=array.shape[iteration_axis],
                                    interval=5, 
                                    repeat_delay=0, 
                                    blit=False)
        # Visualize or store gif
        if out_path is None : plt.show()
        else : ani.save(out_path, writer='imagemagick', fps=None, dpi=None)
        # Close the matplot
        plt.close()
    # Encode as NumPy file
    elif format == "npy" : np.save(out_path, array)
    # Encode as ITK file
    else:
        array_sitk = sitk.GetImageFromArray(array)
        sitk.WriteImage(array_sitk, out_path)

#-----------------------------------------------------#
#               XAI Heatmap Visualizer                #
#-----------------------------------------------------#
def visualize_heatmap(image, heatmap, overlay=True, out_path=None, alpha=0.4):
    """ Simple wrapper function to visualize a heatmap encoded as NumPy matrix with a
        [0-1] range as image/volume via matplotlib.

    ??? abstract "Reference - Implementation"
        Author: François Chollet <br>
        Date: April 26, 2020 <br>
        https://keras.io/examples/vision/grad_cam/ <br>

    Args:
        image (numpy.ndarray):          NumPy matrix containing an image or volume.
        heatmap (numpy.ndarray):        NumPy matrix containing a XAI heatmap.
        out_path (str):                 Path in which image is stored (else live output).
        alpha (float):                  Transparency value for heatmap overlap on image (range: [0-1]).
    """
    # Grayscale normalization if required
    if image.shape[-1] == 1 and (np.min(image) < 0 or np.max(image) > 255):
        image = Standardize(mode="grayscale").transform(image)
    elif image.shape[-1] != 1 and (np.min(image) < 0 or np.max(image) > 255):
        raise ValueError("Visualizer does not support multi-channel " + \
                         "non-RGB formats." + \
                         " Array min: " + str(np.min(image)) + \
                         " Array max: " + str(np.max(image))) 
    # If image is grayscale, convert to RGB
    if image.shape[-1] == 1 : image = np.concatenate((image,)*3, axis=-1)
    # Rescale heatmap to grayscale range
    heatmap = np.uint8(heatmap * 255)
    # Overlay the heatmap on the image 
    if overlay:
        # Use jet colormap to colorize heatmap
        jet = cm.get_cmap("jet")
        # Use RGB values of the colormap
        jet_colors = jet(np.arange(256))[:,:3]
        jet_heatmap = jet_colors[heatmap] * 255
        # Superimpose the heatmap on original image
        final_img = jet_heatmap * alpha + (1-alpha) * image
    # Output just the heatmap
    else : final_img = heatmap
    # Apply corresponding 2D visualizer 
    if len(image.shape) == 3 : visualize_image(final_img, out_path=out_path)
    elif len(image.shape) == 4 : visualize_volume(final_img, out_path=out_path)


#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import os
# AUCMEDI Libraries
from aucmedi.xai.methods import xai_dict
from aucmedi.utils.visualizer import *
from aucmedi.data_processing.subfunctions import Resize

#-----------------------------------------------------#
#                    XAI - Decoder                    #
#-----------------------------------------------------#
def xai_decoder(data_gen, model, preds=None, method="gradcam", layerName=None,
                overlay=True, alpha=0.25, preprocess_overlay=True, out_path=None):
    """ XAI Decoder function for automatic computation of Explainable AI heatmaps.

    This module allows to visualize which regions were crucial for the neural network model
    to compute a classification on the provided unknown images.

    - If `out_path` parameter is None, heatmaps are returned as NumPy array.
    - If a path is provided as `out_path`, then heatmaps are stored to disk as PNG files.

    ???+ info "XAI Methods"
        The XAI Decoder can be run with different XAI methods as backbone.

        A list of all implemented methods and their keys can be found here: <br>
        [aucmedi.xai.methods][]

    ??? info "Parameter: preprocess_overlay"
        The XAI method computation is based on the fully preprocessed image.
        However, sometimes it is needed to map the resulting XAI map to the original image.
        
        Subfunctions which drastically alter the image resolution like cropping lead to 
        an incorrect mapping process which is why a slight preprocessing of the images, on
        which the XAI heatmap is overlayed, is recommended.

    ???+ example "Example"
        ```python
        # Create a DataGenerator for data I/O
        datagen = DataGenerator(samples[:3], "images_xray/", labels=None, resize=(299, 299))

        # Get a model
        model = NeuralNetwork(n_labels=3, channels=3, architecture="Xception",
                               input_shape=(299,299))
        model.load("model.xray.keras")

        # Make some predictions
        preds = model.predict(datagen)

        # Compute XAI heatmaps via Grad-CAM (resulting heatmaps are stored in out_path)
        xai_decoder(datagen, model, preds, method="gradcam", out_path="xai.xray_gradcam")
        ```

    Args:
        data_gen (DataGenerator):           A data generator which will be used for inference.
        model (NeuralNetwork):              Instance of a AUCMEDI neural network class.
        preds (numpy.ndarray):              NumPy Array of classification prediction encoded as OHE (output of a AUCMEDI prediction).
        method (str):                       XAI method class instance or index. By default, GradCAM is used as XAI method.
        layerName (str):                    Layer name of the convolutional layer for heatmap computation. If `None`, the last conv layer is used.
        overlay (bool):                     Switch deciding if XAI heatmap should be plotted as overlap on the original image.
                                            If `False`, only the XAI heatmap will be stroed.
        alpha (float):                      Transparency value for heatmap overlap plotting on input image (range: [0-1]).
        preprocess_overlay (bool):          Switch for Subfunction application on visualization. Only relevant if heatmaps are saved to disk.
        out_path (str):                     Output path in which heatmaps are saved to disk as provided `image_format` (DataGenerator).

    Returns:
        images (numpy.ndarray):             Combined array of images. Will be only returned if `out_path` parameter is `None`.
        heatmaps (numpy.ndarray):           Combined array of XAI heatmaps. Will be only returned if `out_path` parameter is `None`.
    """
    # Initialize & access some variables
    batch_size = data_gen.batch_size
    n_classes = model.n_labels
    sample_list = data_gen.samples
    # Prepare XAI output methods
    res_img = []
    res_xai = []
    if out_path is not None and not os.path.exists(out_path) : os.mkdir(out_path)
    # Initialize xai method
    if isinstance(method, str) and method in xai_dict:
        xai_method = xai_dict[method](model.model, layerName=layerName)
    else : xai_method = method

    # Iterate over all samples
    for i in range(0, len(sample_list)):
        # Load overlay image
        if preprocess_overlay:
            img_org = data_gen.preprocess_image(i, 
                                                run_resize=False, 
                                                run_aug=False, 
                                                run_standardize=False)
            shape_org = img_org.shape[0:-1]
        # Load original image
        else:
            img_org = data_gen.sample_loader(sample_list[i], 
                                    data_gen.path_imagedir,
                                    image_format=data_gen.image_format,
                                    grayscale=data_gen.grayscale,
                                    **data_gen.kwargs)
            shape_org = img_org.shape[0:-1]

        # Load processed image
        img_prc = data_gen.preprocess_image(i, run_aug=False)
        img_batch = np.expand_dims(img_prc, axis=0)
        # If preds given, compute heatmap only for argmax class
        if preds is not None:
            ci = np.argmax(preds[i])
            xai_map = xai_method.compute_heatmap(img_batch, class_index=ci)
            xai_map = Resize(shape=shape_org).transform(xai_map)
            postprocess_output(sample_list[i], img_org, xai_map, 
                               n_classes, data_gen, res_img, res_xai, 
                               overlay, out_path, alpha)
        # If no preds given, compute heatmap for all classes
        else:
            sample_maps = []
            for ci in range(0, n_classes):
                xai_map = xai_method.compute_heatmap(img_batch, class_index=ci)
                xai_map = Resize(shape=shape_org).transform(xai_map)
                sample_maps.append(xai_map)
            sample_maps = np.array(sample_maps)
            postprocess_output(sample_list[i], img_org, sample_maps, 
                               n_classes, data_gen, res_img, res_xai, 
                               overlay, out_path, alpha)
    # Return output directly if no output path is defined
    if out_path is None : return res_img, res_xai

#-----------------------------------------------------#
#          Subroutine: Output Postprocessing          #
#-----------------------------------------------------#
""" Helper/Subroutine function for XAI Decoder.

Caches heatmap for direct output or generates a visualization as PNG.
"""
def postprocess_output(sample, image, xai_map, n_classes, data_gen,
                       res_img, res_xai, overlay, out_path, alpha):
    # Update result lists for direct output
    if out_path is None:
        res_img.append(image)
        res_xai.append(xai_map)
    # Generate XAI heatmap visualization
    else:
        # Create XAI path
        if data_gen.image_format:
            xai_file = sample + "." + data_gen.image_format
        else : xai_file = sample
        if os.sep in xai_file : xai_file = xai_file.replace(os.sep, ".")
        path_xai = os.path.join(out_path, xai_file)
        # If preds given (xai_map), output only argmax class heatmap
        if len(image.shape) == len(xai_map.shape)+1 :
            visualize_heatmap(image, xai_map, overlay=overlay,
                              out_path=path_xai, alpha=alpha)
        # If no preds given (sample_maps), output heatmaps for all classes
        else:
            for c in range(0, n_classes):
                if data_gen.image_format is None:
                    ff = len(xai_file.split(".")[-1]) + 1
                elif data_gen.image_format == ".gz" : ff = 7
                else : ff = len(data_gen.image_format) + 1
                path_xai_c = path_xai[:-ff] + ".class_" + str(c) + \
                             path_xai[-ff:]
                visualize_heatmap(image, xai_map[c], overlay=overlay,
                                  out_path=path_xai_c, alpha=alpha)
